#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
from io import BytesIO
from PIL import Image, ImageFilter

# import interfaces
from pyams_file.interfaces import IImage, IThumbnailer, IThumbnailGeometry, IThumbnail

# import packages
from pyams_utils.adapter import ContextAdapter, adapter_config
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty


WEB_FORMATS = ('JPEG', 'PNG', 'GIF')


@implementer(IThumbnailGeometry)
class ThumbnailGeometrry(object):
    """Image thumbnail geometry"""

    x1 = FieldProperty(IThumbnailGeometry['x1'])
    y1 = FieldProperty(IThumbnailGeometry['y1'])
    x2 = FieldProperty(IThumbnailGeometry['x2'])
    y2 = FieldProperty(IThumbnailGeometry['y2'])

    def __repr__(self):
        return '<ThumbnailGeometry: x1,y1={0.x1},{0.y1} - x2,y2={0.x2},{0.y2}>'.format(self)


@adapter_config(context=IImage, provides=IThumbnailer)
class ImageThumbnailer(ContextAdapter):
    """Image thumbnailer adapter"""

    def get_default_geometry(self, options=None):
        """Default thumbnail geometry"""
        geometry = ThumbnailGeometrry()
        width, height = self.context.get_image_size()
        geometry.x1 = 0
        geometry.y1 = 0
        geometry.x2 = width
        geometry.y2 = height
        return geometry

    def create_thumbnail(self, thumbnail_name, format=None):
        # check thumbnail name
        if isinstance(thumbnail_name, str):
            width, height = tuple((int(x) for x in thumbnail_name.split('x')))
        elif isinstance(thumbnail_name, tuple):
            width, height = thumbnail_name
        else:
            return None
        image = Image.open(self.context.get_blob(mode='c'))
        # check format
        if not format:
            format = image.format
        format = format.upper()
        if format not in WEB_FORMATS:
            format = 'JPEG'
        # check image mode
        if image.mode == 'P':
            image = image.convert('RGBA')
        # generate thumbnail
        new_image = BytesIO()
        image.resize((width, height), Image.ANTIALIAS) \
             .filter(ImageFilter.SHARPEN) \
             .save(new_image, format)
        return new_image, format.lower()


@adapter_config(name='square', context=IImage, provides=IThumbnailer)
class ImageSquareThumbnailer(ContextAdapter):
    """Image square thumbnail adapter"""

    def get_default_geometry(self, options=None):
        """Default square thumbnail geometry"""
        geometry = ThumbnailGeometrry()
        width, height = self.context.get_image_size()
        if width >= height:
            geometry.x1 = round((width / 2) - (height / 2))
            geometry.y1 = 0
            geometry.x2 = round((width / 2) + (height / 2))
            geometry.y2 = height
        else:
            geometry.x1 = 0
            geometry.y1 = round((height / 2) - (width / 2))
            geometry.x2 = width
            geometry.y2 = round((height / 2) + (width / 2))
        return geometry

    def create_thumbnail(self, thumbnail_name, format=None):
        geometry = IThumbnail(self.context).get_thumbnail_geometry(thumbnail_name)
        image = Image.open(self.context.get_blob(mode='c'))
        # get thumbnail size
        if isinstance(thumbnail_name, str):
            if ':' in thumbnail_name:
                thumbnailer_name, size = thumbnail_name.split(':', 1)
                width, height = tuple((int(x) for x in size.split('x')))
            else:
                width, height = tuple((int(x) for x in thumbnail_name.split('x')))
        elif isinstance(thumbnail_name, tuple):
            width, height = thumbnail_name
        else:
            return None
        # check format
        if not format:
            format = image.format
        format = format.upper()
        if format not in WEB_FORMATS:
            format = 'JPEG'
        # check image mode
        if image.mode == 'P':
            image = image.convert('RGBA')
        # generate thumbnail
        new_image = BytesIO()
        image.crop((geometry.x1, geometry.y1, geometry.x2, geometry.y2)) \
             .resize((width, height), Image.ANTIALIAS) \
             .filter(ImageFilter.SHARPEN) \
             .save(new_image, format)
        return new_image, format.lower()


@adapter_config(name='pano', context=IImage, provides=IThumbnailer)
class ImagePanoThumbnailer(ContextAdapter):
    """Image panoramic thumbnail adapter"""

    def get_default_geometry(self, options=None):
        """Default panoramic thumbnail geometry"""
        geometry = ThumbnailGeometrry()
        width, height = self.context.get_image_size()
        pano_max_height = width * 9 / 16
        if pano_max_height >= height:
            # image wider
            pano_width = height * 16 / 9
            geometry.x1 = round((width / 2) - (pano_width / 2))
            geometry.y1 = 0
            geometry.x2 = round((width / 2) + (pano_width / 2))
            geometry.y2 = height
        else:
            pano_height = pano_max_height
            geometry.x1 = 0
            geometry.y1 = round((height / 2) - (pano_height / 2))
            geometry.x2 = width
            geometry.y2 = round((height / 2) + (pano_height / 2))
        return geometry

    def create_thumbnail(self, thumbnail_name, format=None):
        geometry = IThumbnail(self.context).get_thumbnail_geometry(thumbnail_name)
        image = Image.open(self.context.get_blob(mode='c'))
        # get thumbnail size
        if isinstance(thumbnail_name, str):
            if ':' in thumbnail_name:
                thumbnailer_name, size = thumbnail_name.split(':', 1)
                width, height = tuple((int(x) for x in size.split('x')))
            else:
                width, height = tuple((int(x) for x in thumbnail_name.split('x')))
        elif isinstance(thumbnail_name, tuple):
            width, height = thumbnail_name
        else:
            return None
        # check aspect ratio
        thumb_size = abs(geometry.x2 - geometry.x1), abs(geometry.y2 - geometry.y1)
        w_ratio = 1. * width / thumb_size[0]
        h_ratio = 1. * height / thumb_size[1]
        ratio = min(w_ratio, h_ratio)
        # check format
        if not format:
            format = image.format
        format = format.upper()
        if format not in WEB_FORMATS:
            format = 'JPEG'
        # check image mode
        if image.mode == 'P':
            image = image.convert('RGBA')
        # generate thumbnail
        new_image = BytesIO()
        image.crop((geometry.x1, geometry.y1, geometry.x2, geometry.y2)) \
             .resize((round(ratio * thumb_size[0]), round(ratio * thumb_size[1])), Image.ANTIALIAS) \
             .filter(ImageFilter.SHARPEN) \
             .save(new_image, format)
        return new_image, format.lower()
